--[[
	need to:
	
	Disable everything if the character is not level 20.
	Trigger scanning at the start of a BG, instead of by slash command.
	Decide when data is to be shared (start of a bg would be good, or have characters declare when they go AFK so other AFK chars can send then).
	When merging data, update the first seen date if it's earlier than the one stored.
	
	Store addon version in savedSettings, to avoid merging incompatible databases.
	Allow for database sections for different character levels, so other brackets may use the addon.
--]]

local equipmentSlots = {
	"HeadSlot",
	"NeckSlot",
	"ShoulderSlot",
	"ChestSlot",
	"WaistSlot",
	"LegsSlot",
	"FeetSlot",
	"WristSlot",
	"HandsSlot",
	"Finger0Slot", --don't need to check both, as it's 'what can fit in this slot', not 'what's in this slot'.
	"Trinket0Slot", --ditto
	"BackSlot",
	"MainHandSlot",
	"SecondaryHandSlot", --unfortunately this means one-hand items end up showing twice because they can go in either hand.  Doesn't matter, just means they get sent to the DB twice, and there's an overwrite of the first.
}

local slotNames = {
	"Head",
	"Neck",
	"Shoulder",
	"Back",
	"Chest",
	"Wrist",
	"Hands",
	"Waist",
	"Legs",
	"Feet",
	"Finger",
	"Trinket",
	"Ranged",
	"Two-Hand",
	"One-Hand",
	"Main Hand",
	"Off Hand",
	"Held In Off-Hand",
}

local armorTypes = {
	"Cloth",
	"Leather",
	"Mail",
	"Plate",
}

local excludeItems = {
	"Fishing Pole", --excludes all fishing poles
	"Mining Pick",
	"Skinning Knife",
	"Burgy Blackheart's Handsome Hat", --instead of listing everything here, might be better to just exclude items with no stats (not counting armor)
}

local statMatchStrings = {
	["Armor"] = "^(%d+) Armor$",
	["Stamina"] = "^+(%d+) Stamina$",
	["Strength"] = "^+(%d+) Strength$",
	["Agility"] = "^+(%d+) Agility$",
	["Intellect"] = "^+(%d+) Intellect$",
	["Spirit"] = "^+(%d+) Spirit$",
	["Crit"] = "^+(%d+) Critical Strike$",
	["Haste"] = "^+(%d+) Haste$",
	["Dodge"] = "^+(%d+) Dodge$",
	["BonusArmor"] = "^+(%d+) Bonus Armor$",
	["Resilience"] = "^+(%d+) PVP Resilience$",
	["FireResist"] = "^+(%d+) Fire Resistance$",
	["ShadowResist"] = "^+(%d+) Shadow Resistance$",
	["SpellPower"] = "^+(%d+) Spell Power$",
	["AttackPower"] = "^+(%d+) Attack Power$",
	["Speed"] = "^Speed ([%.%d]+)$",
	["DPS"] = "^%(([%.%d]+) damage per second%)$",
	["DamageLow"] = "^(%d+)[%s%-%d]- [%s%w]-Damage$", --[%s%w]- to handle damage types still used on wands
	["DamageHigh"] = "^[%s%-%d]-(%d+) [%s%w]-Damage$",
	["ItemID"] = "^itemID=(%d+)$", --not an official one, but we append it onto data pulled from the tooltip, and use this to format it for output
	["Quality"] = "^quality=(%d+)$",--ditto
	--on use
	--on equip
	--chance on hit
}

local stats = {		--for the output functions
	"ItemID",
	"Armor",
	"Stamina",
	"Strength",
	"Agility",
	"Intellect",
	"Spirit",
	"Crit",
	"Haste",
	"Dodge",
	"BonusArmor",
	"Resilience",
	"FireResist",
	"ShadowResist",
	"SpellPower",
	"AttackPower",
	"Speed",
	"DPS",
	"DamageLow",
	"DamageHigh",
}


local equippableItems = {}
local equipmentIn = {}
local equipmentProcessing = {}



if not EquipmentScale_Variables then
	EquipmentScale_Variables = {}
end



-- create the tootip frame that we'll be reading data from
local tip = CreateFrame("GameTooltip", "ScaledEquipment")
local tooltipLines = {}
local line = 0
for i = 1, 20 do
	local left, right = tip:CreateFontString(), tip:CreateFontString()
	left:SetFontObject(GameFontNormal)
	right:SetFontObject(GameFontNormal)

	tip:AddFontStrings(left, right)

	line = line + 1
	tooltipLines[line] = left

	line = line + 1
	tooltipLines[line] = right
end
tip:SetOwner(UIParent, "ANCHOR_NONE") --attach it to the main UI frame, but give it no location so it isn't visible



--hash function to provide different table indices for the same items with different stats, should bliz change things
local function StringHash(text)
	local counter = 1
	local len = string.len(text)
	for i = 1, len, 3 do 
		counter = math.fmod(counter*8161, 4294967279) +  -- 2^32 - 17: Prime!
		(string.byte(text,i)*16776193) +
		((string.byte(text,i+1) or (len-i+256))*8372226) +
		((string.byte(text,i+2) or (len-i+256))*3932164)
	end
	return math.fmod(counter, 4294967291) -- 2^32 - 5: Prime (and different from the prime in the loop)
end



--sorts the item data table before converting it into a string, so that hashing will be consistent, given that Lua key=value tables aren't consistently ordered
local function tableToString(tbl)
	
	local str = ""
	local orderedTable = {}
	
	for k, v in pairs(tbl) do
		table.insert(orderedTable, {key=k,value=v})
	end
	table.sort(orderedTable,function(a, b) return a.key < b.key end)
	for i, v in ipairs (orderedTable) do
		str = str..v.key.."="..v.value.."::"
	end
	return str
end



--format the data and save it out to savedSettings
local function saveEquipment()
	for i, v in pairs(equipmentIn) do
		equipmentProcessing = {}
		equipmentProcessing["Name"] = equipmentIn[i][1]
		for j,w in pairs(equipmentIn[i]) do
			print(tostring(w))
			for statName, matchString in pairs(statMatchStrings) do
				if string.match(w, matchString) then
					equipmentProcessing[statName] = string.match(w, matchString)
				end
			end
			if tContains(armorTypes,w) then
				equipmentProcessing["ArmorType"] = w
			elseif tContains(slotNames,w) then
				equipmentProcessing["Slot"] = w
			end
		end
		local hash = StringHash(tableToString(equipmentProcessing))
		EquipmentScale_Variables[hash] = {}
		EquipmentScale_Variables[hash]["Item"] = equipmentProcessing
		EquipmentScale_Variables[hash]["LastSeenOnDate"] = date("%y%m%d")
		if not EquipmentScale_Variables[hash]["FirstSeenOnDate"] then
			EquipmentScale_Variables[hash]["FirstSeenOnDate"] = date("%y%m%d")
		end
		local _, playerFaction = UnitFactionGroup("player")
		if not ( playerFaction == "Neutral" ) then
			if not EquipmentScale_Variables[hash]["Faction"] then
				EquipmentScale_Variables[hash]["Faction"] = playerFaction
			elseif (  ( EquipmentScale_Variables[hash]["Faction"] == "Alliance" ) and ( playerFaction == "Horde" ) or ( EquipmentScale_Variables[hash]["Faction"] == "Horde" ) and ( playerFaction == "Alliance" )  )then
				EquipmentScale_Variables[hash]["Faction"] = "Both"
			end
		end
	end
end



--dump all equippable items into equipmentIn{}
local function scanEquipped()
	local inTableIndex = 0
	local iQuality
	for i, v in pairs(equipmentSlots) do
		local id, texture, checkRelic = GetInventorySlotInfo(v)
		equippableItems = {} --clear the table
		GetInventoryItemsForSlot(id, equippableItems) --function populates the table equippableItems{} only with the items for slot 'id' that can be equipped on this character, and are currently in your inventory.  For some reason this includes crafted items, that require a skill level, even if you don't have the skill.
		for locationBitstring, itemID in pairs(equippableItems) do
			local player, bank, bags, voidstorage, slot, bag = EquipmentManager_UnpackLocation(locationBitstring)
			tip:ClearLines()
			if player and (not bags) then
				tip:SetInventoryItem("player", slot)
			elseif player and bags then
				tip:SetBagItem(bag, slot)
			else
				break
			end
			equipmentIn[inTableIndex] = {}
			local excludeItem = false
			for c = 1, #tooltipLines do
				local text = tooltipLines[c]:GetText()  -- what GetText() returns doesn't seem to be a purely text string, and attempting to concatenate it with strings was causing WoW to crash to desktop.  Other functions are working fine though.
				if text then
					equipmentIn[inTableIndex][c]=text
					--set flag if it's an item in the above list, or an engineering item that can't be trained for on an f2p
					if tContains(excludeItems,text) or ( string.match(text, "^Requires Engineering %((%d+)%)$") and tonumber(string.match(text, "^Requires Engineering %((%d+)%)$")) > 100 ) then
						excludeItem = true
					end
				end
			end
			_,_,iQuality, _,_,_,_,_,_,_,_ = GetItemInfo(itemID)
			if tostring(iQuality) == "0" or tostring(iQuality) == "1" then --exclude grey and white items
				excludeItem = true
			end
			if not excludeItem then
				table.insert(equipmentIn[inTableIndex], "itemID="..itemID)
				table.insert(equipmentIn[inTableIndex], "quality="..iQuality)
				inTableIndex = inTableIndex + 1
			else
				equipmentIn[inTableIndex] = nil
			end
		end
	end
	saveEquipment()
end



SLASH_ESCALE1 = "/es"
SlashCmdList["ESCALE"] = function(txt)
	scanEquipped()
end

SLASH_ESCALEOUT1 = "/eso"
SlashCmdList["ESCALEOUT"] = function(txt)
	if not EquipmentScaleFrame:IsVisible() then
		EquipmentScaleFrame:Show()
		EquipmentScaleOutputFrame:Show()
		EquipmentScaleScrollFrame:Hide()
		escale_output()
	else
		EquipmentScaleFrame:Hide()
		EquipmentScaleOutputFrame:Hide()
		EquipmentScaleScrollFrame:Hide()
	end
end

SLASH_ESCALEOUTCSV1 = "/esoc"
SlashCmdList["ESCALEOUTCSV"] = function(txt)
	if not EquipmentScaleFrame:IsVisible() then
		EquipmentScaleFrame:Show()
		EquipmentScaleScrollFrame:Show()
		EquipmentScaleOutputFrame:Hide()
		escale_output_csv()
	else
		EquipmentScaleFrame:Hide()
		EquipmentScaleScrollFrame:Hide()
		EquipmentScaleOutputFrame:Hide()
	end
end

---[[


function escale_showTooltip(...)
	local frame, linkdata, link = ...
	if string.match(linkdata or "", "^item:") then
		local iID = string.match(linkdata, "^item:(%d+):")
		ShowUIPanel(GameTooltip)
		GameTooltip:SetOwner(frame, "ANCHOR_CURSOR")
		GameTooltip:SetItemByID(iID)
		GameTooltip:Show()
	end
end

function escale_hideTooltip(...)
    HideUIPanel(GameTooltip)
end
--]]

--code below dumps DB into the output window
---[[
	



function escale_output() --in game table output with item links to show tooltips from.
	local output = {}
	local line = ""
	local longestName = 0

	--get longest item name
	for k, v in pairs(EquipmentScale_Variables) do
		if #EquipmentScale_Variables[k]["Item"]["Name"] > longestName then
			longestName = #EquipmentScale_Variables[k]["Item"]["Name"]
		end
	end

	table.insert(output, "Name" .. string.rep(" ", longestName - 4) .. "  |ID   |Ar |Sm |Sr |Ag |In |Sp |Cr |Ha |Do |BA ||Re |FR |SR |SP |AP |SPD |DPS    |DL  |DH  |")

	table.insert(output, string.rep("-", longestName) .. "--|-----|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|----|-------|----|----|")

	for k, v in pairs(EquipmentScale_Variables) do
		local a = EquipmentScale_Variables[k]["Item"]
		local iQualityColor = "ffaaaaaa"
		if a["Quality"] then
			_,_,_,iQualityColor = GetItemQualityColor(a["Quality"])
		end
	
		local link = "|c".. iQualityColor .. "|Hitem:" .. a["ItemID"] ..":0:0:0:0:0:0:0:20:0:0|h[" .. a["Name"] .. "]|h|r"

		line = link .. string.rep(" ", longestName - #a["Name"])
		for i, x in ipairs(stats) do
			local matchedStat = false
			for j, w in pairs(a) do
				if j == x then
					if x == "ItemID" then
						line = line .. "||" .. string.rep(" ", 5 - #a[x]) .. a[x]
						---[[
					elseif x == "Speed" then
						line = line .. "||" .. string.rep(" ", 4 - #a[x]) .. a[x]
					elseif x == "DPS" then
						line = line .. "||" .. string.rep(" ", 7 - #a[x]) .. a[x]
					elseif x == "DamageLow" then
						line = line .. "||" .. string.rep(" ", 4 - #a[x]) .. a[x]
					elseif x == "DamageHigh" then
						line = line .. "||" .. string.rep(" ", 4 - #a[x]) .. a[x] .. "|"
					else
						line = line .. "||" .. string.rep(" ", 3 - #a[x]) .. a[x]
						--]]
					end
					matchedStat = true
				end
			end
			if not matchedStat then
				if x == "ItemID" then
					line = line .. "||    "
					---[[
				elseif x == "Speed" then
					line = line .. "||    "
				elseif x == "DPS" then
					line = line .. "||       "
				elseif x == "DamageLow" then
					line = line .. "||    "
				elseif x == "DamageHigh" then
					line = line .. "||    ||"
				else
					line = line .. "||   "
					--]]
				end
			end
		end
		table.insert(output, tostring(line))
		table.insert(output, string.rep("-", longestName) .. "--|-----|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|----|-------|----|----|")
	end

	EquipmentScaleOutputFrame:SetHyperlinksEnabled(1)
	EquipmentScaleOutputFrame:Clear()
	EquipmentScaleOutputFrame:SetInsertMode("TOP")
	
	local elem = ""
	for i = #output, 1, -1 do --reverse the order of the table, because we're adding lines to the top, not the bottom, in order to have the text placed at the top of the window.
		elem = output[i]
		EquipmentScaleOutputFrame:AddMessage(elem)
	end
end
--]]



function escale_output_csv() --in game table output to editbox for cop/paste in CSV format
	
	local output = ""
	local line = ""

	output = "Name"
	
	for i, x in ipairs(stats) do
		output = output .."," .. x
	end
	output = output .. "\n"

	for k, v in pairs(EquipmentScale_Variables) do
		local a = EquipmentScale_Variables[k]["Item"]
		line = a["Name"]
		for i, x in ipairs(stats) do
			if a[x] then
				line = line .. "," .. a[x]
			else
				line = line ..","
			end
		end
		output = output..line.."\n"
	end
	EquipmentScaleScrollFrameEditBox:SetText("")
	EquipmentScaleScrollFrameEditBox:Insert(output)
end